@@ -28,6 +28,7 @@ static PyStructSequence_Field flow_id_ntuple_fields[] = {
     {"type", "One of FLOW_TYPE_TCP or FLOW_TYPE_UDP"},
     {NULL}
 };
+
 static PyStructSequence_Desc flow_id_ntuple_desc = {
     "fparser.FlowId",
     NULL,
@@ -116,11 +117,11 @@ static PyObject *fparser_stop(PyObject *self, PyObject *args) {
     return Py_None;
 }
 
-static void fparser_dealloc(struct fparser *fparser) {
+static void fparser_dealloc2(struct fparser *fparser, int free_object) {
     Py_BEGIN_ALLOW_THREADS
-    
+
     kill_parser(fparser->parser, 1);
-    
+
     Py_END_ALLOW_THREADS
 
     free(fparser->parser);
@@ -131,7 +132,23 @@ static void fparser_dealloc(struct fparser *fparser) {
     Py_XDECREF(fparser->udp_flow_callback);
     Py_XDECREF(fparser->after_kill_callback);
 
-    fparser->ob_type->tp_free((PyObject *) fparser);
+    if (free_object) {
+        fparser->ob_type->tp_free((PyObject *) fparser);
+    }
+}
+
+static void fparser_dealloc(struct fparser *fparser) {
+    /*
+      There is a dangling reference to fparser in fparser->parser->user.
+      When the dealloc method is called fparser does not exist anymore
+      in Python (its refcount is 0). Make sure the dangling reference
+      gets cleared. All subsequent calls from callbacks to get the
+      fparser objects should fail.
+     */
+    fparser->parser->user = NULL;
+
+
+    fparser_dealloc2(fparser, 1);
 }
 
 static PyObject *fparser_new(PyTypeObject *type, PyObject *args, PyObject *kwds) {
@@ -192,6 +209,11 @@ static PyObject *fparser_new(PyTypeObject *type, PyObject *args, PyObject *kwds)
     self->udp_flow_callback = NULL;
 
     init_parser(self->parser);
+
+    /* We need to be able to perform async callbacks from C to Python
+       and hence we need to cache the fparser Python object somewhere
+       in the C struct. This creates a nasty dangling reference. Be
+       sure to handle properly. */
     self->parser->user = self;
 
     init_logging(self->parser, log_level);
@@ -242,7 +264,7 @@ static PyObject *fparser_new(PyTypeObject *type, PyObject *args, PyObject *kwds)
 
 fail:
 
-    fparser_dealloc(self);
+    fparser_dealloc2(self, 0);
     return NULL;
 }
 
@@ -447,7 +469,7 @@ static void fparser_tcp_flow_dealloc(struct fparser_tcp_flow *fparser_flow) {
         kh_del(tcp_in_python, table, k);
     }
 
-    Py_DECREF(fparser_flow->fparser);
+    Py_CLEAR(fparser_flow->fparser);
     fparser_flow->ob_type->tp_free((PyObject *) fparser_flow);
 }
 
@@ -466,7 +488,7 @@ static void fparser_udp_flow_dealloc(struct fparser_udp_flow *fparser_flow) {
         kh_del(udp_in_python, table, k);
     }
 
-    Py_DECREF(fparser_flow->fparser);
+    Py_CLEAR(fparser_flow->fparser);
     fparser_flow->ob_type->tp_free((PyObject *) fparser_flow);
 }
 
@@ -520,21 +542,14 @@ static PyObject *fparser_get_flow_iter(PyObject *self,
         return NULL;
     }
 
-    iter = PyObject_New(struct fparser_flow_iter, &fparser_flow_iter_type);
+    iter = (struct fparser_flow_iter *) PyObject_CallObject((PyObject *) & fparser_flow_iter_type, NULL);
     if (!iter) {
-        PyErr_SetString(FlowParserError, "Unable to allocate iter object");
-        return NULL;
-    }
-
-    if (!PyObject_Init((PyObject *) iter, &fparser_flow_iter_type)) {
-        Py_DECREF(iter);
-        PyErr_SetString(FlowParserError, "Unable to init iter object");
         return NULL;
     }
 
     kv_init(iter->flows);
     iter->curr_pos = 0;
-    
+
     iter->fparser = fparser;
     Py_INCREF(iter->fparser);
 
@@ -1076,22 +1091,17 @@ static struct fparser_tcp_flow *tcp_flow_to_python(struct flow_id *id,
         return py_flow;
     }
 
-    py_flow = PyObject_New(struct fparser_tcp_flow, &fparser_tcp_flow_type);
+    py_flow = PyObject_New(struct fparser_tcp_flow, (PyTypeObject *) & fparser_tcp_flow_type);
     if (!py_flow) {
         return NULL;
     }
 
-    if (!PyObject_Init((PyObject *) py_flow, &fparser_tcp_flow_type)) {
-        Py_DECREF(py_flow);
-        return NULL;
-    }
-
     py_flow->flow = flow;
     py_flow->id = id;
-    
+
     py_flow->fparser = fparser;
     Py_INCREF(py_flow->fparser);
-    
+
     init_tcp_iter(py_flow);
 
     k = kh_put(tcp_in_python, fparser->tcp_cache_table, id, &ret);
@@ -1102,10 +1112,22 @@ static struct fparser_tcp_flow *tcp_flow_to_python(struct flow_id *id,
 
 static void trigger_on_kill_callback(struct parser_base *p) {
     PyGILState_STATE d_gstate;
+
+    if (p->user == NULL) {
+        d_gstate = PyGILState_Ensure();
+
+        PyErr_SetString(PyExc_RuntimeError,
+                "FParser object already dead. Will not perform on kill callback.");
+        PyErr_Print();
+
+        PyGILState_Release(d_gstate);
+        return;
+    }
+
     struct fparser *fparser = (struct fparser *) p->user;
     PyObject *result;
 
-    if (fparser->after_kill_callback == NULL) {
+    if (fparser == NULL || fparser->after_kill_callback == NULL) {
         return;
     }
 
@@ -1136,6 +1158,17 @@ void offload_tcp_flow(struct parser_base *p, struct flow_id *id,
     PyObject *arglist;
     PyObject *result;
 
+    if (p->user == NULL) {
+        d_gstate = PyGILState_Ensure();
+
+        PyErr_SetString(PyExc_RuntimeError,
+                "FParser object already dead. Will not offload TCP flow.");
+        PyErr_Print();
+
+        PyGILState_Release(d_gstate);
+        return;
+    }
+
     struct fparser *fparser = (struct fparser *) p->user;
     if (fparser->tcp_flow_callback == NULL) {
         return;
@@ -1182,22 +1215,17 @@ static struct fparser_udp_flow *udp_flow_to_python(struct flow_id *id,
         return py_flow;
     }
 
-    py_flow = PyObject_New(struct fparser_udp_flow, &fparser_udp_flow_type);
+    py_flow = PyObject_New(struct fparser_udp_flow, (PyTypeObject *) & fparser_udp_flow_type);
     if (!py_flow) {
         return NULL;
     }
 
-    if (!PyObject_Init((PyObject *) py_flow, &fparser_udp_flow_type)) {
-        Py_DECREF(py_flow);
-        return NULL;
-    }
-
     py_flow->flow = flow;
     py_flow->id = id;
-    
+
     py_flow->fparser = fparser;
     Py_INCREF(py_flow->fparser);
-    
+
     init_udp_iter(py_flow);
 
     k = kh_put(udp_in_python, fparser->udp_cache_table, id, &ret);
@@ -1223,6 +1251,17 @@ void offload_udp_flow(struct parser_base *parser,
     PyObject *result;
     struct fparser_udp_flow *py_flow;
 
+    if (parser->user == NULL) {
+        d_gstate = PyGILState_Ensure();
+
+        PyErr_SetString(PyExc_RuntimeError,
+                "FParser object already dead. Will not offload UDP flow.");
+        PyErr_Print();
+
+        PyGILState_Release(d_gstate);
+        return;
+    }
+
     struct fparser *fparser = (struct fparser *) parser->user;
     if (fparser->udp_flow_callback == NULL) {
         return;
