@@ -27,7 +27,7 @@ void init_parser(struct parser_base *p) {
     p->debug_level = LOG_ERR;
     p->flow_timeout = 15 * 60 * (uint64_t) SEC;
     p->tcp_afterfin_timeout = 30 * (uint64_t) SEC;
-    
+
     p->collector_status = NOT_STARTED;
     p->dumper_status = NOT_STARTED;
     p->sniffer_status = NOT_STARTED;
@@ -43,10 +43,10 @@ void close_parser(struct parser_base *p) {
 
         clear_flows(p);
         kh_destroy(32, p->flows_table);
-        
+
         chunk_base_close(tcp_chunks, &p->tcp_chunks);
         chunk_base_close(udp_chunks, &p->udp_chunks);
-        
+
         p->flows_table = NULL;
     } else {
         INFO(p, "Flow table already closed");
@@ -54,16 +54,16 @@ void close_parser(struct parser_base *p) {
 }
 
 void kill_parser(struct parser_base *p, int dealloc_parser) {
-    p->to_kill = 1;
-    
+    __sync_val_compare_and_swap(&p->to_kill, 0, 1);
+
     close_sniff(p);
     close_collector(p);
     close_dumper(p);
-    
+
     if (dealloc_parser) {
         close_parser(p);
     }
-    
+
     if (p->after_kill_callback) {
         p->after_kill_callback(p);
         p->after_kill_callback = NULL;
@@ -201,12 +201,12 @@ void clear_flows(struct parser_base *p) {
 
 void find_flow_and_id(const struct parser_base *p, const struct flow_id *id,
         struct flow_tuple *tuple) {
-    int is_missing;    
+    int is_missing;
     khiter_t k;
 
     k = kh_get(32, p->flows_table, id);
     is_missing = (k == kh_end(p->flows_table));
-    
+
     if (is_missing) {
         tuple->flow = NULL;
         tuple->flow_id = NULL;
@@ -318,7 +318,7 @@ struct flow *lookup_flow(struct parser_base *p,
         const struct flow_id *id) {
     struct flow *flow = NULL;
     int ret;
-    
+
     /* Initially we will only get a read lock as it is more likely 
      * that we won't need to construct a new flow (there are a lot more 
      * packets than flows) */
@@ -341,7 +341,7 @@ struct flow *lookup_flow(struct parser_base *p,
         ret = pthread_rwlock_wrlock(&p->flows_table_lock);
         if (ret == 0) {
             DEBUG(p, "Acquired write lock on flows table");
-            
+
             /* We managed to get a write lock on the primary table. Now we 
              * call get_flow to create/get a new flow. Note that it is 
              * possible that the table has changed and the flow has already 
@@ -370,7 +370,7 @@ struct flow *lookup_flow(struct parser_base *p,
         }
         DEBUG(p, "Released read lock on flows table");
     }
-    
+
     return flow;
 }
 
@@ -471,6 +471,14 @@ static inline int unlock_flow_mutex(struct parser_base *p, struct flow *flow) {
     return 0;
 }
 
+static inline void update_parser_bookkeeping(struct parser_base *p, uint64_t new_timestamp) {
+    int res;
+    __sync_fetch_and_add(&p->pkts_this_second, 1);
+    do {
+        res = __sync_bool_compare_and_swap(&p->last_rx_time, p->last_rx_time, new_timestamp);
+    } while (!res);
+}
+
 /**
  * Adds a new TCP packet to the parser. This function accepts libpcap-style 
  * structs and handles locking/unlocking the flow. 
@@ -521,8 +529,7 @@ int handle_tcp_pkt(struct parser_base *p,
     }
 
     // some bookkeeping
-    p->pkts_this_second++;
-    p->last_rx_time = timestamp;
+    update_parser_bookkeeping(p, timestamp);
     flow->last_rx_time = timestamp;
     flow->expiry_time = timestamp + p->flow_timeout;
     flow->pkts_this_period++;
@@ -618,8 +625,7 @@ int handle_udp_pkt(struct parser_base *p,
     }
 
     // some bookkeeping
-    p->pkts_this_second++;
-    p->last_rx_time = timestamp;
+    update_parser_bookkeeping(p, timestamp);
     flow->last_rx_time = timestamp;
     flow->expiry_time = timestamp + p->flow_timeout;
     flow->pkts_this_period++;
